from collections import OrderedDict

from django.conf import settings


class SearchSlicer(object):
    """We want to search things like a seesaw. take that mike parent.
    """
    def __init__(self, *args, **kwargs):
        self.default_queryset = None
        self.querysets = OrderedDict()
        self.index = 0
        self.limit = kwargs.get("limit", getattr(settings, "READING_LIST_LIMIT", 100))

    def __iter__(self):
        return self

    def next(self):
        if self.index >= self.limit:
            raise StopIteration
        for validator, queryset in self.querysets.items():
            if validator(self.index):
                try:
                    result = queryset.next()
                    self.index += 1
                    return result
                except StopIteration:
                    pass
        result = self.default_queryset.next()
        self.index += 1
        return result

    def __next__(self):
        return self.next()

    def register_queryset(self, queryset, validator=None, default=False):
        """
        Add a given queryset to the iterator with custom logic for iteration.

        :param queryset: List of objects included in the reading list.
        :param validator: Custom logic to determine a queryset's position in a reading_list.
            Validators must accept an index as an argument and return a truthy value.
        :param default: Sets the given queryset as the primary queryset when no validator applies.
        """
        if default or self.default_queryset is None:
            self.default_queryset = queryset
            return
        if validator:
            self.querysets[validator] = queryset
        else:
            raise ValueError(
                """Querysets require validation logic to integrate with reading lists."""
            )


def FirstSlotSlicer(primary_query, secondary_query, limit=30):  # noqa
    """
    Inject the first object from a queryset into the first position of a reading list.

    :param primary_queryset: djes.LazySearch object. Default queryset for reading list.
    :param secondary_queryset: djes.LazySearch object. first result leads the reading_list.
    :return list: mixed reading list.
    """
    reading_list = SearchSlicer(limit=limit)
    reading_list.register_queryset(primary_query)
    reading_list.register_queryset(secondary_query, validator=lambda x: bool(x == 0))
    return reading_list



from django.views.decorators.cache import cache_control
from django.shortcuts import get_object_or_404

from rest_framework import views, status
from rest_framework.permissions import AllowAny
from rest_framework.response import Response

from bulbs.content.models import Content
from bulbs.recirc.serializers import RecircContentSerializer


class RecircViewSet(views.APIView):

    permission_classes = (AllowAny,)

    def get(self, request, pk):
        content = get_object_or_404(Content, id=pk)
        if not content.is_published:
            return Response(status=status.HTTP_404_NOT_FOUND)

        serializer = RecircContentSerializer(
            content.get_recirc_content(count=3).execute(),
            many=True
        )

        return Response(serializer.data, status=status.HTTP_200_OK)


class InlineRecircViewSet(views.APIView):

    permission_classes = (AllowAny,)

    def get(self, request, pk):
        content = get_object_or_404(Content, id=pk)
        if not content.is_published:
            return Response(status=status.HTTP_404_NOT_FOUND)

        serializer = RecircContentSerializer(
            content.get_inline_recirc_content(count=3).execute(),
            many=True
        )

        return Response(serializer.data, status=status.HTTP_200_OK)


inline_recirc = cache_control(max_age=600)(InlineRecircViewSet.as_view())
recirc = cache_control(max_age=600)(RecircViewSet.as_view())



from rest_framework.routers import DefaultRouter

from .viewsets import PollViewSet, AnswerViewSet

router = DefaultRouter(trailing_slash=True)

api_v1_router = DefaultRouter()
api_v1_router.register(
    r"poll",
    PollViewSet,
    base_name="poll")

api_v1_router.register(
    r"poll-answer",
    AnswerViewSet,
    base_name="answer")

urlpatterns = api_v1_router.urls



from django.conf import settings
from django.conf.urls import url, include

from bulbs.cms_notifications.api import notifications_view
from .views import api_v1_router, MeViewSet


urlpatterns = (
    url(r"^me/logout/?$", "django.contrib.auth.views.logout", name="logout"),
    url(r"^me/?$", MeViewSet.as_view({"get": "retrieve"}), name="me"),
    url(r"^", include(api_v1_router.urls))  # noqa
)

if "bulbs.promotion" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^", include("bulbs.promotion.urls")),
    )

if "bulbs.special_coverage" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^", include("bulbs.special_coverage.urls")),
    )

if "bulbs.cms_notifications" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^notifications/(?P<pk>\d+)?", notifications_view, name="notifications"),
    )

if "bulbs.contributions" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^contributions/", include("bulbs.contributions.urls")),
    )

if "bulbs.campaigns" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^", include("bulbs.campaigns.urls")),
    )

if "bulbs.sections" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^", include("bulbs.sections.urls")),
    )

if "bulbs.poll" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^", include("bulbs.poll.api")),
    )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('contributions', '0003_auto_20151103_1147'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contributorrole',
            name='payment_type',
            field=models.IntegerField(choices=[(0, 'Flat Rate'), (1, 'FeatureType'), (2, 'Hourly'), (3, 'Manual')], default=3),
        ),
        migrations.AlterField(
            model_name='rate',
            name='name',
            field=models.IntegerField(null=True, choices=[(0, 'Flat Rate'), (1, 'FeatureType'), (2, 'Hourly'), (3, 'Manual'), (4, 'Override')]),
        ),
    ]



import contextdecorator
import json
import logging
import random
import string
import time

from elasticsearch_dsl.connections import connections

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.utils import timezone

from djes.apps import indexable_registry
from djes.management.commands.sync_es import get_indexes, sync_index

from model_mommy import mommy
from rest_framework.test import APIClient
from six import PY2

from bulbs.content.models import Content


def make_content(*args, **kwargs):
    if "make_m2m" not in kwargs:
        kwargs["make_m2m"] = True

    if len(args) == 1:
        klass = args[0]
    else:
        models = indexable_registry.families[Content]
        model_keys = []
        for key in models.keys():
            if key not in ['content_content', 'poll_poll']:
                model_keys.append(key)
        key = random.choice(model_keys)
        klass = indexable_registry.all_models[key]

    content = mommy.make(klass, **kwargs)
    return content


def random_title():
    return ''.join([random.choice(string.ascii_uppercase) for _ in range(10)])


class JsonEncoder(json.JSONEncoder):
    def default(self, value):
        """Convert more Python data types to ES-understandable JSON."""
        iso = _iso_datetime(value)
        if iso:
            return iso
        if PY2 and isinstance(value, str):
            return unicode(value, errors='replace')  # TODO: Be stricter.
        if isinstance(value, set):
            return list(value)
        return super(JsonEncoder, self).default(value)


def _iso_datetime(value):
    """
    If value appears to be something datetime-like, return it in ISO format.

    Otherwise, return None.
    """
    if hasattr(value, 'strftime'):
        if hasattr(value, 'hour'):
            return value.isoformat()
        else:
            return '%sT00:00:00' % value.isoformat()


class BaseIndexableTestCase(TestCase):
    """A TestCase which handles setup and teardown of elasticsearch indexes."""

    elasticsearchLogger = logging.getLogger('elasticsearch')

    def setUp(self):
        """ If you're reading this. I am gone and dead, presumably.
            Elasticsearch's logging is quite loud and lets us know
            about anticipated errors, so I set the level to ERROR only.
            If elasticsearch is giving you trouble in tests and you
            aren't seeing any info, get rid of this. God bless you.
        """
        self.elasticsearchLogger.setLevel(logging.ERROR)
        self.es = connections.get_connection("default")
        self.indexes = get_indexes()

        self.now = timezone.now()

        for index in list(self.indexes):
            self.es.indices.delete_alias("{}*".format(index), "_all", ignore=[404])
            self.es.indices.delete("{}*".format(index), ignore=[404])

        for index, body in self.indexes.items():
            sync_index(index, body)

        self._wait_for_allocation()

    def _wait_for_allocation(self):
        """Wait for shards to be ready, to avoid flaky test errors when ES searches triggered before
        cluster is initialized.
        This is especially important for tests that do not trigger any sort of ES refresh.
        """
        MAX_WAIT_SEC = 30
        start = time.time()
        while (time.time() - start) < MAX_WAIT_SEC:
            if all(len(shard) and shard[0]['state'] == 'STARTED'
                   for shard in self.es.search_shards()['shards']):
                return
        self.fail('One or more ES shards failed to startup within {} seconds'.format(MAX_WAIT_SEC))

    def tearDown(self):
        for index in list(self.indexes):
            self.es.indices.delete_alias("{}*".format(index), "_all", ignore=[404])
            self.es.indices.delete("{}*".format(index), ignore=[404])


class BaseAPITestCase(BaseIndexableTestCase):
    """A base test case, allowing tearDown and setUp of the ES index"""

    def setUp(self):
        super(BaseAPITestCase, self).setUp()
        User = get_user_model()
        admin = self.admin = User.objects.create_user("admin", "tech@theonion.com", "secret")
        admin.is_staff = True
        admin.save()
        self.api_client = APIClient()
        self.api_client.force_authenticate(user=admin)
        # reverse("content-detail")

    def give_permissions(self):
        publish_perm = Permission.objects.get(codename="publish_content")
        change_perm = Permission.objects.get(codename="change_content")
        promote_perm = Permission.objects.get(codename="promote_content")
        self.admin.user_permissions.add(publish_perm, change_perm, promote_perm)

    def give_author_permissions(self):
        publish_perm = Permission.objects.get(codename="publish_own_content")
        self.admin.user_permissions.add(publish_perm)

    def remove_permissions(self):
        self.admin.user_permissions.clear()





from django import template

register = template.Library()


@register.inclusion_tag('special_coverage/bulbs_sc_landing.html', takes_context=True)
def special_coverage_landing_partial(context):
    return context



import logging

from django.db import models, transaction

from djes.models import Indexable

from bulbs.content.models import Content, ElasticsearchImageField

from .mixins import AnswerMixin, PollMixin


logger = logging.getLogger(__name__)


"""
Sodahead API Integration Rationale

The sodahoad api implements a POST and a PUT endpoint.

In the payload to these endpoints, answers are keyed

answer_0n      => "Question 1 Text"
answer_0n+1    => "Question 2 Text"
answer_0n+...  => "Question 3 Text"

Or however many answers you would like.

Sodahead provides no way to re-order/delete these answers.

So everytime we create an answer, we want to increment
the associated answer key.

This is especially tricky, as if we have the following
answers:

In the context of a single Poll:

poll.models.Poll(pk=1)

poll.models.Answer(pk=1) => answer_01
poll.models.Answer(pk=2) => answer_02
poll.models.Answer(pk=3) => answer_03

And we delete Answer(pk=2)

The next answer needs to be

poll.models.Answer(pk=5) => answer_04

So we cannot merely add up the answers of a poll, we
have to keep track of how many answers we have already
created.

This is done by the field `last_answer_index` on the Poll model.

It is incremented by 1 every time an Answer is saved.
"""


class Poll(Content, PollMixin):

    # This keeps Poll out of Content.search_objects
    class Mapping(Content.Mapping):
        poll_image = ElasticsearchImageField()

        class Meta():
            orphaned = True


class Answer(Indexable, AnswerMixin):
    poll = models.ForeignKey(
        Poll,
        on_delete=models.CASCADE,
        related_name='answers'
    )

    class Mapping:
        answer_image = ElasticsearchImageField()

    def save(self, *args, **kwargs):
        # using transaction/select_for_update here because we don't want to
        # run this block in parallel. If we did, we would end up with answers
        # with the same value for `answer.sodahead_answer_id`.
        # This is because we are deriving sodahead_answer_id based on the count
        # of answers that have ever been created for an poll.
        #   (see top of file for in depth explainer)
        with transaction.atomic():
            poll = Poll.objects.select_for_update().get(pk=self.poll_id)
            if self.sodahead_answer_id:
                super(Answer, self).save(*args, **kwargs)
            else:
                # See Rationale at top of file for in depth explanation
                # of what is going on here. We have to track how many answers
                # have ever been created for a Poll.
                # A simple count will not suffice because we dont want to
                # re-use any answer keys.
                answer_index = poll.last_answer_index + 1
                self.sodahead_answer_id = 'answer_%02d' % (answer_index)
                super(Answer, self).save(*args, **kwargs)
                poll.last_answer_index += 1
                poll.save()



from django.contrib import admin

from .models import FreelanceProfile


class FreelanceProfileAdmin(admin.ModelAdmin):
    fields = ("payroll_name", "is_freelance", "is_manager")
    list_display = ("get_full_name",)

    def get_queryset(self, request):
        qs = super(FreelanceProfileAdmin, self).get_queryset(request)
        return qs.distinct()

    def get_full_name(self, obj):
        return obj.contributor.get_full_name()
    get_full_name.admin_order_field = 'contributor'


admin.site.register(FreelanceProfile, FreelanceProfileAdmin)



import hashlib
import os

from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from rest_framework.authtoken.models import Token


User = get_user_model()


class Command(BaseCommand):

    help = "Creates a new user & app token for a client request."

    def add_arguments(self, parser):
        parser.add_argument(
            '--username',
            dest='username',
            help='The username for the given application.',
            required=True,
            type=str
        )

    def generate_random_password(self, length=13):
        random_data = os.urandom(128)
        return hashlib.md5(random_data).hexdigest()[:16]

    def handle(self, *args, **kwargs):
        username = kwargs.get('username')
        user, created = User.objects.get_or_create(username=username, defaults={
            'password': self.generate_random_password(),
            'is_staff': True
        })
        token = Token.objects.create(user=user)
        self.stdout.write(token.key)



from django.conf import settings
from django.contrib.auth import get_user_model

from .models import Contribution, ContributorRole, FreelanceProfile, FEATURETYPE


def merge_roles(dominant_name, deprecated_name):
    """
    Merges a deprecated role into a dominant role.
    """
    dominant_qs = ContributorRole.objects.filter(name=dominant_name)
    if not dominant_qs.exists() or dominant_qs.count() != 1:
        return
    dominant = dominant_qs.first()
    deprecated_qs = ContributorRole.objects.filter(name=deprecated_name)
    if not deprecated_qs.exists() or deprecated_qs.count() != 1:
        return
    deprecated = deprecated_qs.first()

    # Update Rates
    if not dominant.flat_rates.exists() and deprecated.flat_rates.exists():
        flat_rate = deprecated.flat_rates.first()
        flat_rate.role = dominant
        flat_rate.save()

    if not dominant.hourly_rates.exists() and deprecated.hourly_rates.exists():
        hourly_rate = deprecated.hourly_rates.first()
        hourly_rate.role = dominant
        hourly_rate.save()

    for ft_rate in deprecated.feature_type_rates.all():
        dom_ft_rate = dominant.feature_type_rates.filter(feature_type=ft_rate.feature_type)

        if dom_ft_rate.exists() and dom_ft_rate.first().rate == 0:
            dom_ft_rate.first().delete()

        if not dom_ft_rate.exists():
            ft_rate.role = dominant
            ft_rate.save()

    # Update contributions
    for contribution in deprecated.contribution_set.all():
        contribution.role = dominant
        contribution.save()

    # Update overrides
    for override in deprecated.overrides.all():
        dom_override_qs = dominant.overrides.filter(contributor=override.contributor)
        if not dom_override_qs.exists():
            override.role = dominant
            override.save()
        else:
            dom_override = dom_override_qs.first()
            for flat_override in override.override_flatrate.all():
                flat_override.profile = dom_override
                flat_override.save()
            for hourly_override in override.override_hourly.all():
                hourly_override.profile = dom_override
                hourly_override.save()
            for feature_type_override in override.override_feature_type.all():
                feature_type_override.profile = dom_override
                feature_type_override.save()


def get_forced_payment_contributions(start_date, end_date, qs=None):
    if not qs:
        qs = Contribution.objects.filter(force_payment=True)
    include = qs.filter(
        payment_date__range=(start_date, end_date)
    ) | qs.filter(
        payment_date__isnull=True
    )
    exclude = qs.exclude(
        payment_date__isnull=True
    ).exclude(
        payment_date__range=(start_date, end_date),
    )
    return include, exclude


def update_content_contributions(content, author):
    if not Contribution.objects.filter(content=content, contributor=author).exists():
        role_name = getattr(settings, 'DEFAULT_CONTRIBUTOR_ROLE', 'default')
        role, created = ContributorRole.objects.get_or_create(name=role_name)
        if created:
            role.payment_type = FEATURETYPE
            role.save()
        Contribution.objects.create(content=content, contributor=author, role=role)


def import_payroll_names(lookup_string):
    user = get_user_model()
    lookup_string = lookup_string.decode('unicode_escape').encode('ascii', 'ignore')
    rows = lookup_string.split('\n')
    rows = [row.rstrip('\n').rstrip('\r') for row in rows]
    rows = [row.split('\t') for row in rows]
    for row in rows:
        full_name, payroll_name = row[0], row[1]
        ordered_name = full_name.strip().split(' ')
        if len(ordered_name) == 2:
            first_name, last_name = ordered_name[0], ordered_name[1]
            qs = user.objects.filter(first_name=first_name, last_name=last_name)
            if qs.count() != 1:
                # print 'Cannot find user: {}'.format(full_name)
                return
            else:
                user = qs.first()
                profile = getattr(user, 'freelanceprofile', None)
                if not profile:
                    profile = FreelanceProfile.objects.create(
                        contributor=user,
                        is_freelance=True
                    )
                else:
                    profile.payroll_name = payroll_name.strip()
                    profile.save()



"""Module to generate and send a report to contributors containing a log of their contributions."""
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.mail import EmailMultiAlternatives
from django.template import loader
from django.utils import timezone


User = get_user_model()


# Define constants.
CONTRIBUTION_SETTINGS = getattr(settings, "CONTRIBUTIONS", {})
EMAIL_SETTINGS = CONTRIBUTION_SETTINGS.get("EMAIL", {})
DEFAULT_SUBJECT = "Contribution Report."
TEMPLATE = "reporting/__contribution_report.html"


class EmailReport(object):
    """Generate an email report for contributors."""

    def __init__(self, **kwargs):
        # if "start" in kwargs:
        self.now = timezone.now()
        self.month = kwargs.get("month", self.now.month)
        self.year = kwargs.get("year", self.now.year)

        self._contributors = []
        self._deadline = kwargs.get("deadline")
        self._start = kwargs.get("start")
        self._end = kwargs.get("end")

    def send_contributor_email(self, contributor):
        """Send an EmailMessage object for a given contributor."""
        body = self.get_email_body(contributor)
        mail = EmailMultiAlternatives(
            subject=EMAIL_SETTINGS.get("SUBJECT", DEFAULT_SUBJECT),
            from_email=EMAIL_SETTINGS.get("FROM"),
            bcc=EMAIL_SETTINGS.get("BCC"),
            headers={"Reply-To": EMAIL_SETTINGS.get("REPLY_TO")}
        )
        mail.attach_alternative(body, "text/html")
        if EMAIL_SETTINGS.get("ACTIVE", False):
            mail.to = [contributor.email]
        else:
            mail.to = EMAIL_SETTINGS.get("TO")
        mail.send()

    def send_mass_contributor_emails(self):
        """Send report email to all relevant contributors."""
        # If the report configuration is not active we only send to the debugging user.
        if EMAIL_SETTINGS.get("ACTIVE", False):
            for contributor in self.contributors:
                self.send_contributor_email(contributor)
        else:
            for email in EMAIL_SETTINGS.get("TO", []):
                self.send_contributor_email(User.objects.get(email=email))

    def get_email_body(self, contributor):
        contributions = self.get_contributions_by_contributor(contributor)
        total = sum([contribution.pay for contribution in contributions if contribution.pay])
        contribution_types = {}
        for contribution in contributions:
            contribution_types[contribution] = contribution.content._meta.concrete_model.__name__
        context = {
            "content_type": contribution.content._meta.concrete_model.__name__,
            "contributor": contributor,
            "contributions": contribution_types,
            "deadline": self.deadline,
            "total": total
        }
        return loader.render_to_string(TEMPLATE, context)

    def get_contributors(self):
        """Return a list of contributors with contributions between the start/end dates."""
        return User.objects.filter(
            contributions__content__published__gte=self.start,
            contributions__content__published__lt=self.end
        ).distinct()

    def get_contributions_by_contributor(self, contributor, **kwargs):
        """Return a list of all contributions associated with a contributor for a given month."""
        return contributor.contributions.filter(
            content__published__gte=self.start, content__published__lt=self.end
        )

    @property
    def contributors(self):
        """Property to retrieve or access the list of contributors."""
        if not self._contributors:
            self._contributors = self.get_contributors()
        return self._contributors

    @property
    def deadline(self):
        """Set deadline to next day if no deadline provided."""
        if not self._deadline:
            self._deadline = self.now + timezone.timedelta(days=1)
        return self._deadline

    @property
    def start(self):
        if not self._start:
            self._start = timezone.datetime(day=1, month=self.month, year=self.year)
        return self._start

    @property
    def end(self):
        if not self._end:
            next_month = (self.start.month + 1) % 12
            year = self.start.year
            if next_month == 1:
                year += 1
            elif next_month == 0:
                next_month = 12
            self._end = timezone.datetime(day=1, month=next_month, year=year)
        return self._end



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import djbetty.fields


class Migration(migrations.Migration):

    dependencies = [
        ('special_coverage', '0005_auto_20151103_1358'),
    ]

    operations = [
        migrations.AddField(
            model_name='specialcoverage',
            name='_image_alt',
            field=models.CharField(max_length=255, null=True, editable=False, blank=True),
        ),
        migrations.AddField(
            model_name='specialcoverage',
            name='_image_caption',
            field=models.CharField(max_length=255, null=True, editable=False, blank=True),
        ),
        migrations.AddField(
            model_name='specialcoverage',
            name='image',
            field=djbetty.fields.ImageField(default=None, alt_field=b'_image_alt', null=True, caption_field=b'_image_caption', blank=True),
        ),
    ]



from django.conf import settings
from django.http import Http404
from django.views.generic import DetailView
from django.views.decorators.cache import cache_control

import requests


class SeriesDetailView(DetailView):

    def get_template_names(self):
        template_names = ["videos/series-detail.html"]
        return template_names

    def get(self, request, *args, **kwargs):
        slug = kwargs.get("slug", None)

        if not slug:
            raise Http404

        response = requests.get(settings.VIDEOHUB_BASE_URL+"/series/"+slug+".json")

        if not response.ok:
            raise Http404

        self.object = response.json()
        context = self.get_context_data()
        response = self.render_to_response(context)

        return response

    def get_context_data(self, *args, **kwargs):
        context = super(SeriesDetailView, self).get_context_data()

        context["channel_name"] = self.object['channel_name']
        context["series_name"] = self.object['series_name']
        context["series_description"] = self.object['series_description']
        context["total_seasons"] = self.object['total_seasons']
        context["total_episodes"] = self.object['total_episodes']

        return context

series_detail = cache_control(max_age=600)(SeriesDetailView.as_view())
